"""Provides with QueryRunner.
"""
from pymnesia.core.composition import composite
from pymnesia.core.entities.entity import Entity


class QueryRunner:
    __slots__ = [
        "__entity_class",
        "__unit_of_work",
    ]

    def __init__(self, entity_class, unit_of_work):
        self.__entity_class = entity_class
        self.__unit_of_work = unit_of_work

    def __entities(self):
        """Returns the entities in the unit of work matching the entity class.

        :return: A list of entities.
        """
        return [value for key, value in getattr(self.__unit_of_work, self.__entity_class.__tablename__).items()]

    def __run_query_funcs(self, *query_funcs):
        """Runs the query functions passed when fetch methods are called.

        :param query_funcs: The query functions to run.
        :return: A list of results where filter, order by functions may have been run.
        """
        compose_query_funcs = composite(*query_funcs)

        return list(compose_query_funcs(self.__entities()))

    def __run_or_funcs(self, *or_func_groups):
        """Runs the query functions passed when fetch methods are called.

        :param query_funcs: The query functions to run.
        :return: A list of results where filter, order by functions may have been run.
        """
        or_results = []
        for or_functions in or_func_groups:
            compose_or_funcs = composite(*or_functions)
            or_results.extend(list(compose_or_funcs(self.__entities())))

        return or_results

    def __load_relations(self, entity: Entity):
        """Loads an entity relations from the unit of work.

        :param entity: The entity for which to load the relations.
        :return: None
        """
        for relation_name, relation in self.__entity_class.__conf__.relations.items():
            if relation.is_owner:
                relation_key_value = getattr(entity, relation.key)
                if relation_key_value is not None:
                    if relation.relation_type == "one_to_one":
                        setattr(
                            entity,
                            relation_name,
                            getattr(
                                self.__unit_of_work,
                                relation.entity_cls_resolver.__tablename__
                            )[getattr(entity, relation.key)]
                        )
                    if relation.relation_type == "one_to_many":
                        relation_keys = getattr(entity, relation.key)
                        relations = []
                        for relation_key in relation_keys:
                            relations.append(
                                getattr(
                                    self.__unit_of_work,
                                    relation.entity_cls_resolver.__tablename__
                                )[relation_key]
                            )
                        setattr(
                            entity,
                            relation_name,
                            relations
                        )

    def fetch(self, *args, or_function_groups: list, order_by_functions: list, limit: int) -> list:
        """Returns multiple results based on a series of parameters,
        such as a where clause, a limit, and order_by, etc...

        :param args: The query functions to run.
        :param or_function_groups: The or function groups to run.
        :param order_by_functions: The order by function groups to run.
        :param limit: The limit to use.
        :return: A list of entities
        """
        results = self.__entities()
        if args:
            results = self.__run_query_funcs(*args)
            [  # pylint: disable=expression-not-assigned
                results.append(r) for r in self.__run_or_funcs(*or_function_groups)
                if r not in results
            ]
        if order_by_functions:
            compose_order_by_funcs = composite(*order_by_functions)
            results = compose_order_by_funcs(results)
        if limit:
            results = results[0:limit]
        for result in results:
            self.__load_relations(result)

        return results

    def fetch_one(self, *args, or_function_groups: list, order_by_functions: list):
        """Returns the first result of a query based on a series of parameters,
        such as a where clause, an order_by, etc...

        :param args: The query functions to run.
        :param or_function_groups: The query functions to run.
        :param order_by_functions: The order by function groups to run.
        :return: A single entity
        """
        results = self.__entities()
        if args:
            results = self.__run_query_funcs(*args)
        results += self.__run_or_funcs(*or_function_groups)
        if order_by_functions:
            compose_order_by_funcs = composite(*order_by_functions)
            results = compose_order_by_funcs(results)
        result = results[0]

        self.__load_relations(entity=result)

        return result



"""Provides with a metaclass for UnitOfWork.
"""
from pymnesia.core.common.originator_interface import OriginatorInterface
from pymnesia.core.entities.registry import DEFAULT_E_CLASSES_REGISTRY


class UnitOfWorkMeta(type(OriginatorInterface)):

    # noinspection PyMethodParameters
    def __new__(mcs, name, bases, attrs):
        attrs["__slots__"] = ["__state", "__replica"]
        for entity_cls_resolver in DEFAULT_E_CLASSES_REGISTRY.all_configs():  # pylint: disable=unused-variable
            attrs["__slots__"].append(entity_cls_resolver.__tablename__)

        # noinspection PyTypeChecker
        return super().__new__(mcs, name, bases, attrs)



"""Provides with a base class for entities.
"""
from pymnesia.core.entities.meta import EntityMeta


class DeclarativeBase(metaclass=EntityMeta):
    """Base class for entities.

    @DynamicAttrs"""

    def __init__(self, **kwargs):
        pass



"""Provides with command related APIs.
"""
from pymnesia.core.transaction.transaction import InMemoryTransaction
from pymnesia.core.unit_of_work.unit_of_work import UnitOfWork

__all__ = [
    "transaction",
]


def transaction(unit_of_work: UnitOfWork) -> InMemoryTransaction:
    """Returns a transaction for a given unit of work.

    :param unit_of_work: The unit of work for which to create a transaction.
    :return: A new InMemoryTransaction.
    """
    return InMemoryTransaction(originator=unit_of_work)



"""Provides with a class to store an entity class configuration.
"""
from dataclasses import dataclass, field
from typing import Dict


@dataclass()
class EntityClsConf:
    relations: Dict = field(default_factory=lambda: {})



"""Provides with an exception raised when a primary key was not declared.
"""


class MissingPrimaryKeyException(Exception):
    def __init__(self):
        super().__init__("Entity was declared without a primary key.")



"""Provides with Query class.
"""
import re
from typing import Callable, Generator
from functools import partial

from pymnesia.core.entities.entity import Entity
from pymnesia.core.query.filter.registry import find_filter_function
from pymnesia.core.query.functions import order_by
from pymnesia.core.query.runner import QueryRunner
from pymnesia.core.query.filter.functions import *  # pylint: disable=wildcard-import,unused-wildcard-import
from pymnesia.core.query.filter.relations import *  # pylint: disable=wildcard-import,unused-wildcard-import

FIELD_OPERATOR_REGEX = re.compile(r'^(?P<field>\w+)::(?P<operator>\w+)$')
RELATION_PROPERTY_REGEX = re.compile(r'^(?P<rel>\w+)\.(?P<rel_property>.*)$')

__all__ = [
    "Query"
]


class Query:
    """Allows to store a query parameters and run the query using a QueryRunner.
    """

    __slots__ = [
        "__entity_class",
        "__unit_of_work",
        "__query_runner",
        "__limit",
        "__query_functions",
        "__or_functions",
        "__order_by_functions",
    ]

    def __init__(self, entity_class, unit_of_work):
        self.__entity_class = entity_class
        self.__unit_of_work = unit_of_work
        self.__query_runner = QueryRunner(
            entity_class=self.__entity_class,
            unit_of_work=self.__unit_of_work
        )
        self.__limit = 0
        self.__query_functions = []
        self.__or_functions = []
        self.__order_by_functions = []

    def fetch(self) -> Generator[Entity, None, None]:
        """Returns multiple results based on a series of parameters,
        such as a where clause, a limit, and order_by, etc...

        :return: A list of entities
        """
        for entity in self.__query_runner.fetch(
            *self.__query_functions,
            or_function_groups=self.__or_functions,
            order_by_functions=self.__order_by_functions,
            limit=self.__limit
        ):
            yield entity

    def fetch_one(self):
        """Returns the first result of a query based on a series of parameters,
        such as a where clause, an order_by, etc...

        :return: A single entity
        """
        return self.__query_runner.fetch_one(
            *self.__query_functions,
            or_function_groups=self.__or_functions,
            order_by_functions=self.__order_by_functions,
        )

    def where(self, clause: dict):
        """Processes and stores the parameters for a 'where' clause.
        Each condition in the where clause is matched with a filter function.

        :param clause: The where clause to use for the query.
        :return: The query to use for chaining.
        """
        self.__query_functions.extend(self.__build_filter_composite_funcs(clause))

        return self

    def where_with_composition(self, filter_funcs: list[Callable]):
        """Adds composite functions to be used as conditional filters.

        :param filter_funcs: A list of callables matching the Pymnesia filter API.
        :return: The query to use for chaining.
        """
        self.__query_functions.extend(filter_funcs)

        return self

    def or_(self, clause: dict):
        """Processes and stores parameters for 'or' clauses.

        :param clause: The or clause to use for the query. This clause is to be considered as an 'or and', meaning that
        every parameter passed will be processed as additional filter.
        Eg:
            query().orders().where({"status":"shipped"}).or_({"status":"ready_to_ship", "is_paid": True}).fetch()
            will return all the entities that have a status "shipped" or the ones
            that have a status "ready_to_ship" AND are paid.
        :return: The query to use for chaining.
        """
        self.__or_functions.append(self.__build_filter_composite_funcs(clause))

        return self

    def __build_filter_composite_funcs(self, clause: dict) -> list[Callable]:
        """Builds partial/composite functions to use for where/or clauses.

        :param clause: The clause to build the functions from.
        :return: A list of composite callables.
        """
        filter_funcs = []
        for condition, value in clause.items():
            filter_args = {"value": value}
            matched_condition = re.match(FIELD_OPERATOR_REGEX, condition)
            if matched_condition:
                filter_func = find_filter_function(filter_name=matched_condition.group("operator"))
                filter_args["field"] = matched_condition.group("field")
            else:
                filter_func = find_filter_function(filter_name="eq")
                filter_args["field"] = condition
                matched_rel_property = re.match(RELATION_PROPERTY_REGEX, condition)
                if matched_rel_property:
                    filter_func = find_filter_function(filter_name="eq", relational=True)
                    filter_args["unit_of_work"] = self.__unit_of_work
                    filter_args["field"] = matched_rel_property.group("rel_property")
                    filter_args["relation"] = self.__entity_class.__conf__.relations[matched_rel_property.group("rel")]
            filter_funcs.append(partial(
                filter_func,
                **filter_args
            ))

        return filter_funcs

    def order_by(self, direction: str, key: str):
        """Stores parameters to use for ordering (sorting) a query result by a key.

        :param direction: Whether the result should be ordered by in an ascending or descending manner.
        :param key: The property to use for ordering.
        :return: The query to use for chaining.
        """
        self.__order_by_functions.append(partial(
            order_by,
            direction=direction,
            key=key,
        ))

        return self

    def limit(self, limit: int):
        """Adds a limit to a query.

        :param limit: The limit to set, 0 if not limit.
        :return: The query to use for chaining.
        """
        self.__limit = limit

        return self



from typing import Any, Iterable
from itertools import groupby

from pymnesia.core.entities.relations import Relation
from pymnesia.core.query.filter.registry import register_filter_func

__all__ = ["filter_rel_eq"]


@register_filter_func(filter_name="eq", relational=True)
def filter_rel_eq(entities: Iterable, field: str, value: Any, unit_of_work, relation: Relation) -> filter:
    """Filters entities whose field is equal to a given value.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :param unit_of_work:
    :param relation:
    :return: An iterable (filter) containing the filtered results.
    """
    base_query = getattr(unit_of_work.query(), relation.entity_cls_resolver.__tablename__)()
    rels = base_query.where({field: value}).fetch()

    if relation.relation_type == "one_to_one":
        filtered_rel_ids = [rel.id for rel in rels]

        return filter(lambda e: getattr(e, relation.key) in filtered_rel_ids, entities)

    reverse_foreign_key_name = relation.reverse + "_id"
    grouped_filtered_rels = groupby(rels, key=lambda e: getattr(e, reverse_foreign_key_name))
    filtered_ids = []
    for key, group in grouped_filtered_rels:  # pylint: disable=unused-variable
        filtered_ids.append(key)

    return filter(lambda e: e.id in filtered_ids, entities)



"""Provides with an interface for EntityClassResolver.
"""
from abc import ABC, abstractmethod
from typing import Type

from pymnesia.core.entities.entity import Entity


class EntityClassResolverInterface(ABC):
    @abstractmethod
    def update_entity_cls(self, entity_cls: Type[Entity]):
        """
        :param entity_cls:
        :return:
        """



"""Provides with a registry to store and use entities' configuration.
"""
from typing import Generator

from pymnesia.core.entities.entity_resolver import EntityClassResolver
from pymnesia.core.entities.registry.interface import PymnesiaRegistryInterface


class PymnesiaRegistry(PymnesiaRegistryInterface):
    """Registry class that stores entity configurations.
    """

    def __init__(self):
        self._entries = []

    def register(self, entity_class) -> EntityClassResolver:
        resolver_cls = type(
            entity_class.__name__,
            (EntityClassResolver,),
            {}
        )
        resolver = resolver_cls(
            entity_cls=entity_class,
            registry=self,
        )
        self._entries.append(resolver)

        return resolver

    def unregister(self, entity_cls_resolver):
        self._entries.remove(entity_cls_resolver)

    def all_configs(self) -> Generator[EntityClassResolver, None, None]:
        for resolver in self._entries:
            yield resolver



"""Provides with an in memory transaction.
"""
from typing import Generator, List

from pymnesia.core.common.originator_interface import OriginatorInterface
from pymnesia.core.unit_of_work.memento.base import UnitOfWorkMemento

__all__ = [
    "InMemoryTransaction",
]


class InMemoryTransaction:
    """Caretaker for in memory unit of works."""

    def __init__(self, originator: OriginatorInterface):
        self.__originator = originator
        self.__mementos: List[UnitOfWorkMemento] = [self.__originator.save()]

    def commit(self):
        """
        Commits the current state of the originator
        """
        self.__mementos.append(self.__originator.save())

    def rollback(self):
        *_, last = self.history()
        self.__originator.restore(memento=last)

    def history(self) -> Generator[UnitOfWorkMemento, None, None]:
        """
        Yields every memento stored.
        :return: 
        """
        for memento in self.__mementos:
            yield memento



"""Provides with an interface for PymnesiaRegistry.
"""
from abc import ABC, abstractmethod
from typing import Type

from pymnesia.core.entities.entity import Entity


class PymnesiaRegistryInterface(ABC):
    """Interface for Pymnesia registry.
    """

    @abstractmethod
    def register(self, entity_class: Type[Entity]):
        """Registers an entity class by adding its resolver to the registry.
        The resolver will later be used to manage proper entity class instantiations.

        :param entity_class: The entity class to register.
        :return: None
        """

    @abstractmethod
    def unregister(self, entity_cls_resolver):
        """Disposes of an entity class by removing its resolver from the registry.

        :param entity_cls_resolver: The entity class' resolver
        :return: None
        """

    @abstractmethod
    def all_configs(self):
        """Iterator that returns all the registered entity class resolvers.

        :return: A generator yielding EntityClassResolver instances.
        """



"""Provides with a memento for UnitOfWork.
"""

__all__ = ["UnitOfWorkMemento"]


class UnitOfWorkMemento:
    """Base class for UnitOfWorkMemento.

    @DynamicAttrs"""

    def __init__(self, **kwargs):
        pass



"""Provides with Entity class.
"""


class Entity:
    """Entity class.
    Used as base class when making dataclasses from declarative bases.

    @DynamicAttrs"""

    def __init__(self, *args, **kwargs):
        pass



import datetime
import re
from typing import Union, Any, Iterable

from pymnesia.core.query.filter.registry import register_filter_func

__all__ = ["filter_eq", "filter_not", "filter_greater_than", "filter_greater_than_or_equal", "filter_less_than",
           "filter_less_than_or_equal", "filter_match", "filter_in"]


@register_filter_func(filter_name="eq")
def filter_eq(entities: Iterable, field: str, value: Any) -> filter:
    """
    Filters entities whose field is equal to a given value.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: getattr(e, field) == value, entities)


@register_filter_func(filter_name="not")
def filter_not(entities: Iterable, field: str, value: Any) -> filter:
    """
    Filters entities whose field is not equal to a given value.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: getattr(e, field) != value, entities)


@register_filter_func(filter_name="gt")
def filter_greater_than(entities: Iterable, field: str, value: Union[str, int, type(datetime)]) -> filter:
    """
    Filters entities whose field is greater than a given value.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: getattr(e, field) > value, entities)


@register_filter_func(filter_name="gte")
def filter_greater_than_or_equal(entities: Iterable, field: str, value: Union[str, int, type(datetime)]) -> filter:
    """
    Filters entities whose field is greater than or equal to a given value.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: getattr(e, field) >= value, entities)


@register_filter_func(filter_name="lt")
def filter_less_than(entities: Iterable, field: str, value: Union[str, int, type(datetime)]) -> filter:
    """
    Filters entities whose field is less than a given value.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: getattr(e, field) < value, entities)


@register_filter_func(filter_name="lte")
def filter_less_than_or_equal(entities: Iterable, field: str, value: Union[str, int, type(datetime)]) -> filter:
    """
    Filters entities whose field is less than or equal to a given value.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: getattr(e, field) <= value, entities)


@register_filter_func(filter_name="match")
def filter_match(entities: Iterable, field: str, value: re.Pattern) -> filter:
    """
    Filters entities whose field matches a given value, based on a regular expression, and where re.match is used for
    evaluation.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: re.match(value, getattr(e, field)), entities)


@register_filter_func(filter_name="in")
def filter_in(entities: Iterable, field: str, value: list) -> filter:
    """
    Filters entities whose field's value is in a range of values.
    :param entities: The entities to filter.
    :param field: The field to compare.
    :param value: The value to use for comparison.
    :return: An iterable (filter) containing the filtered results.
    """
    return filter(lambda e: getattr(e, field) in value, entities)



"""Provides with an exception raised when a tablename was not declared.
"""


class MissingTablenameException(Exception):
    def __init__(self):
        super().__init__("Entity was declared without a tablename.")



"""Provides with EntityClassResolver.
"""
from typing import Type

from pymnesia.core.entities.entity import Entity
from pymnesia.core.entities.registry.interface import PymnesiaRegistryInterface
from pymnesia.core.entities.entity_cls_resolver_interface import EntityClassResolverInterface


class EntityClassResolver(EntityClassResolverInterface):
    """Resolver that instantiates the proper entity class when called.
    """

    def __init__(
            self,
            entity_cls: Type[Entity],
            registry: PymnesiaRegistryInterface,
    ):
        self._entity_cls = entity_cls
        self._registry = registry

    def __call__(self, *args, **kwargs):
        return self._entity_cls(*args, **kwargs)

    def __getattr__(self, item):
        return getattr(self._entity_cls, item)

    def update_entity_cls(self, entity_cls: Type[Entity]):
        """Updates the entity class associated with the resolver.

        :param entity_cls: The updated entity class.
        :return: None
        """
        self._entity_cls = entity_cls



"""Provides with the base class for QueryEngine.
"""

__all__ = [
    "QueryEngine",
]


class QueryEngine:
    """Base class for QueryEngine.

    Allows to query a unit of work tables.

    @DynamicAttrs
    """

    def __init__(
            self,
            unit_of_work,
    ):
        self.unit_of_work = unit_of_work



"""Provides with entity class declaration related APIs.
"""
from typing import Type

from pymnesia.core.entities.base import DeclarativeBase

__all__ = [
    "declarative_base",
]


def declarative_base() -> Type[DeclarativeBase]:
    """Returns the base class to declare an entity.

    :return: The declarative base for entity classes.
    """
    return DeclarativeBase



"""Provides with an interface for """
from abc import ABC, abstractmethod

from pymnesia.core.unit_of_work.memento.base import UnitOfWorkMemento


class OriginatorInterface(ABC):

    @abstractmethod
    def save(self) -> UnitOfWorkMemento:
        """Saves the current unit of work and returns a memento"""

    @abstractmethod
    def restore(self, memento: UnitOfWorkMemento):
        """Restores a unit of work states from a previous memento"""



"""Provides with relation fields.
"""
from dataclasses import dataclass, field

from pymnesia.core.entities.entity_cls_resolver_interface import EntityClassResolverInterface


@dataclass()
class Relation:

    reverse: str

    relation_type: str = field(default="one_to_one")

    is_nullable: bool = field(default=True)

    entity_cls_resolver: EntityClassResolverInterface = field(default=None)

    key: str = field(default=None)

    is_owner: bool = field(default=False)



"""Provides with a memento for UnitOfWork.
"""
from dataclasses import make_dataclass, field
from typing import Dict
from uuid import UUID

from pymnesia.core.entities.registry.interface import PymnesiaRegistryInterface
from pymnesia.core.unit_of_work.memento.base import UnitOfWorkMemento

__all__ = [
    "UnitOfWorkMementoMeta",
    "unit_of_work_metaclass",
]


class UnitOfWorkMementoMeta(type):
    """Metaclass for UnitOfWorkMemento.
    """
    _registry: PymnesiaRegistryInterface = None

    def __new__(mcs, name, bases, attrs):  # pylint: disable=unused-argument
        return make_dataclass('UnitOfWorkMemento', make_fields(mcs._registry), bases=(UnitOfWorkMemento,))


def make_fields(entities_registry) -> list:
    """Makes the required fields for UnitOfWorkMemento, based on the registered entities.
    :param entities_registry: The registry to use for build.

    :return: A list of fields.
    """
    fields = [("state", int)]
    fields += [
        (
            entity_cls_resolver.__tablename__,
            Dict[UUID, entity_cls_resolver],
            field(default_factory=lambda: {})  # pylint: disable=invalid-field-call
        ) for entity_cls_resolver in entities_registry.all_configs()
    ]

    return fields


def unit_of_work_metaclass(registry_: PymnesiaRegistryInterface):
    """UnitOfWorkMemento metaclass maker.

    :param registry_:
    :return:
    """
    return type(
        UnitOfWorkMementoMeta.__name__,
        (UnitOfWorkMementoMeta,),
        {"_registry": registry_}
    )



from functools import reduce


def composite(*func):
    """Functional composition"""

    def compose(g, f):  # pylint: disable=invalid-name
        def h(*args, **kwargs):  # pylint: disable=invalid-name
            return g(f(*args, **kwargs), **kwargs)

        return h

    return reduce(compose, func)



"""Provides with a Unit Of Work (storage class that will act as the originator)
"""
import time
from copy import deepcopy

from pymnesia.core.common.originator_interface import OriginatorInterface
from pymnesia.core.query.engine.meta import query_engine_metaclass
from pymnesia.core.unit_of_work.memento.meta import unit_of_work_metaclass
from pymnesia.core.unit_of_work.meta import UnitOfWorkMeta
from pymnesia.core.unit_of_work.memento.base import UnitOfWorkMemento
from pymnesia.core.entities.registry import DEFAULT_E_CLASSES_REGISTRY

__all__ = ["UnitOfWork"]


class UnitOfWork(OriginatorInterface, metaclass=UnitOfWorkMeta):
    """InMemory storage originator

    @DynamicAttrs"""

    def __init__(
            self,
            state: int,
    ):
        self.__state = state
        self.__replica = None
        self.__unit_of_work_memento_cls = unit_of_work_metaclass(registry_=DEFAULT_E_CLASSES_REGISTRY)(
            "UnitOfWorkMemento",
            (UnitOfWorkMemento, ),
            {},
        )
        self.__query_engine_cls = query_engine_metaclass(registry_=DEFAULT_E_CLASSES_REGISTRY)(
            "QueryEngine",
            (),
            {}
        )
        self.__setup()

    def __setup(self):
        self.__replica = self.__unit_of_work_memento_cls(
            state=self.__state,
        )
        for entity_cls_resolver in DEFAULT_E_CLASSES_REGISTRY.all_configs():  # pylint: disable=unused-variable
            setattr(self, entity_cls_resolver.__tablename__, {})

    def save_entity(self, entity):
        """Saves an entity in the replica.

        :param entity: The entity to save.
        """
        self.__replica.state = time.time_ns()
        table = getattr(self.__replica, entity.__tablename__)
        table[entity.id] = entity

    def save(self) -> UnitOfWorkMemento:
        """Saves the unit of work replica's current state.
        Used by the caretaker to commit.

        :return: The memento of the saved state.
        """
        self.__state = deepcopy(self.__replica.state)  # pylint: disable=no-member
        replica_values = {}

        for entity_cls_resolver in DEFAULT_E_CLASSES_REGISTRY.all_configs():  # pylint: disable=unused-variable
            tablename = entity_cls_resolver.__tablename__
            values = deepcopy(getattr(self.__replica, tablename))
            setattr(
                self,
                tablename,
                values
            )
            replica_values[tablename] = values

        memento = self.__unit_of_work_memento_cls(
            state=self.__state,
            **replica_values
        )

        return memento

    def restore(self, memento: UnitOfWorkMemento):
        """Restores the unit of work to a previous state.

        :param memento: The memento to use for restoration.
        :return: None
        """
        self.__replica = memento

    def query(self):
        """Returns a query engine to be used for querying the unit of work current state.

        :return: A query engine instantiated with the current unit of work state.
        """
        return self.__query_engine_cls(unit_of_work=self)



"""Provides with unit of work related APIs.
"""
import time

from pymnesia.core.unit_of_work.unit_of_work import UnitOfWork

__all__ = [
    "uow",
]


def uow():
    """Returns an instance of UnitOfWork.
    A unit of work is the equivalent of a database.
    It stores entities and manages their states and related operations, such as save.

    :return: An instance of UnitOfWork.
    """
    return UnitOfWork(state=time.time_ns())



"""Provides with field related APIs.
"""
from typing import Any, Callable

from pymnesia.core.entities.field import Field, UNDEFINED
from pymnesia.core.entities.relations import Relation

__all__ = [
    "field",
    "relation",
]


def field(
        default: Any = UNDEFINED,
        default_factory: Callable = UNDEFINED,
):
    """Returns a Field instance.
    The instance is to be processed later on by the entity metaclass and various procedural functions.

    :param default: The default value for the field.
    :param default_factory: A callback used to return a default value for the field.
    :return: A Field instance.
    """
    if default is not UNDEFINED:
        return Field(default=default)
    return Field(default_factory=default_factory)


def relation(
        reverse: str,
        is_nullable: bool = True,
) -> Relation:
    """Returns a Relation instance.
    The instance is to be processed later on by the entity metaclass and various procedural functions.

    :param reverse: The name of the reverse relation.
    :param is_nullable: Whether the relation is nullable or not.
    :return: A Relation instance.
    """
    return Relation(
        reverse=reverse,
        is_nullable=is_nullable,
    )



"""Provides with Field class to configure an entity field.
"""
from dataclasses import dataclass, field
from typing import Any, Callable


class UNDEFINED:
    """Provides with a way to avoid defensive code when building fields.
    Similar to MISSING from dataclasses module.
    """


@dataclass()
class Field:
    """Provides with a class to store an entity field configuration.
    Many properties of which will later be used to create a dataclass field.
    """
    default: Any = field(default=UNDEFINED)

    default_factory: Callable = field(default=UNDEFINED)



"""Provides with various functions for querying.
Filter functions are located in a dedicated module.
"""
from typing import Iterable


def order_by(entities: Iterable, direction: str, key: str) -> list:
    """
    Orders (sort) a query result by a key.

    :param entities: The entities to order by.
    :param direction: Whether the result should be ordered by in an ascending or descending way.
    :param key: The property to use for ordering.
    :return: A list of sorted entities.
    """
    return sorted(entities, key=lambda e: getattr(e, key), reverse=direction == "desc")



"""Provides with a metaclass for entities.
"""
from dataclasses import make_dataclass, field, fields as dataclass_fields, MISSING, Field as DataclassField
from typing import List, Type, Tuple, Dict, Union, get_origin, get_args, Any
from uuid import UUID

from pymnesia.core.entities.entity_cls_conf import EntityClsConf
from pymnesia.core.entities.field import UNDEFINED, Field
from pymnesia.core.entities.entity import Entity
from pymnesia.core.entities.registry import DEFAULT_E_CLASSES_REGISTRY
from pymnesia.core.entities.registry.exceptions.missing_primary_key import MissingPrimaryKeyException
from pymnesia.core.entities.registry.exceptions.missing_tablename import MissingTablenameException
from pymnesia.core.entities.relations import Relation
from pymnesia.core.entities.entity_resolver import EntityClassResolver


class EntityMeta(type):
    """Metaclass for entities.

    Responsible for making dataclasses out of declared entities and registering them in the active registry.
    """

    def __new__(mcs, name, base, attrs):
        if name == "DeclarativeBase":
            return super().__new__(mcs, name, base, attrs)

        fields = []
        relation_fields = {}
        relations = {}

        _validate_entity_cls_declaration(attrs=attrs)

        tablename = attrs["__tablename__"]
        conf = EntityClsConf()

        for field_name, field_type in attrs["__annotations__"].items():
            if field_name in attrs.keys():
                field_as_attr = attrs[field_name]
                if isinstance(field_as_attr, Field):
                    fields.append((
                        field_name,
                        field_type,
                        field(  # pylint: disable=invalid-field-call
                            **_extract_entity_field_attrs(field_as_attr, UNDEFINED)
                        )
                    ))
                if isinstance(field_as_attr, Relation):
                    one_to_many_relation = _which_one_to_many_relation(typed_attr=field_type)
                    relation = field_type
                    if one_to_many_relation:
                        relation = one_to_many_relation
                        field_as_attr.relation_type = "one_to_many"
                    relations[relation] = field_as_attr
                    field_as_attr.is_owner = True
                    _add_foreign_key_to_fields(
                        relation_name=field_name,
                        fields=fields,
                        is_nullable=field_as_attr.is_nullable,
                        relation_type=field_as_attr.relation_type,
                    )
                    relation_fields[relation] = field_name
            else:
                one_to_many_relation = _which_one_to_many_relation(typed_attr=field_type)
                if one_to_many_relation:
                    relations[one_to_many_relation] = Relation(
                        reverse=tablename[0:-1],
                        relation_type="one_to_many",
                        is_owner=True,
                    )
                    relation_fields[one_to_many_relation] = field_name
                    _add_foreign_key_to_fields(
                        relation_name=field_name,
                        fields=fields,
                        is_nullable=True,
                        relation_type="one_to_many",
                    )
                elif issubclass(field_type, Entity):
                    relations[field_type] = Relation(reverse=tablename[0:-1], is_owner=True)
                    _add_foreign_key_to_fields(
                        relation_name=field_name,
                        fields=fields,
                        is_nullable=True,
                        relation_type="one_to_one",
                    )
                    relation_fields[field_type] = field_name
                else:
                    fields.append((
                        field_name,
                        field_type
                    ))

        cls_resolver = DEFAULT_E_CLASSES_REGISTRY.register(
            _make_entity_dataclass(
                name=name,
                tablename=tablename,
                fields=fields,
                conf=conf,
            )
        )
        if relations:
            _make_relations(
                relations=relations,
                cls_resolver=cls_resolver,
                relation_fields=relation_fields,
                cls_resolver_current_fields=_extract_entity_cls_fields(cls_resolver)
            )

        return cls_resolver


def empty_list_factory() -> List:
    """Used as an alternative for lambda: [], since lambdas are compared by reference and hence cannot be tested.

    :return: A empty list.
    """
    return []


def _make_relations(
        relations: Dict[EntityClassResolver, Relation],
        cls_resolver: EntityClassResolver,
        relation_fields: Dict[EntityClassResolver, str],
        cls_resolver_current_fields: List[Tuple],
):
    """Makes relations for both the relation owner and the related entities.
    After making dataclasses for every related entity, the relation owner entity is updated with
    the newly made relations.

    :param relations: The relations to make.
    :param cls_resolver: The relation owner entity class resolver.
    :param relation_fields: A dictionary where the key is the relation and the value the relation name.
    :param cls_resolver_current_fields: The current existing fields for the relationship owner.
    :return: None
    """
    tablename = cls_resolver.__tablename__

    for relation, relation_field in relations.items():
        reverse_relation_type = "one_to_one"
        if relation_field.relation_type == "one_to_many":
            reverse_relation_type = "many_to_one"
        relation_current_fields = _extract_entity_cls_fields(relation)
        _add_foreign_key_to_fields(
            relation_name=relation_field.reverse,
            fields=relation_current_fields,
            is_nullable=False,
            relation_type=reverse_relation_type,
        )
        _add_relation_to_fields(
            relation_name=relation_field.reverse,
            relation=cls_resolver,
            fields=relation_current_fields,
            is_nullable=True,
            relation_type=reverse_relation_type,
        )
        relation_new_cls = _make_entity_dataclass(
            name=relation.__name__,
            tablename=relation.__tablename__,
            fields=sorted(
                relation_current_fields,
                key=lambda e: e[2].default == MISSING and e[2].default == MISSING,
                reverse=True,
            ),
            conf=relation.__conf__,
        )
        relation.update_entity_cls(relation_new_cls)
        _add_relation_to_fields(
            relation_name=relation_fields[relation],
            relation=relation,
            fields=cls_resolver_current_fields,
            is_nullable=True,
            relation_type=relation_field.relation_type,
        )
        _add_relation_to_entity_cls_conf(
            conf=cls_resolver.__conf__,
            relation=relation,
            relation_name=relation_fields[relation],
            relation_field=relation_field,
            relation_type=relation_field.relation_type,
        )
        reverse = relation_new_cls.__tablename__[0:-1]
        if relation_field.relation_type == "one_to_many":
            reverse = relation_new_cls.__tablename__
        _add_relation_to_entity_cls_conf(
            conf=relation_new_cls.__conf__,
            relation=cls_resolver,
            relation_name=relation_field.reverse,
            relation_field=Relation(
                reverse=reverse,
                relation_type=reverse_relation_type,
            ),
            relation_type=reverse_relation_type,
        )

    updated_entity_cls = _make_entity_dataclass(
        name=cls_resolver.__name__,
        tablename=tablename,
        fields=cls_resolver_current_fields,
        conf=cls_resolver.__conf__,
    )
    cls_resolver.update_entity_cls(updated_entity_cls)


def _make_entity_dataclass(
        name: str,
        tablename: str,
        fields: List[Tuple],
        conf: EntityClsConf,
) -> Type[Entity]:
    """Makes a dataclass from entity class parameters.

    :param name: The entity class name
    :param tablename: The table name under which the entity should be registered
    :param fields: The fields to create for the entity dataclass
    :param conf:
    :return: A entity dataclass
    """
    entity_cls = make_dataclass(name, fields, bases=(Entity,))
    entity_cls.__tablename__ = tablename
    entity_cls.__conf__ = conf

    # noinspection PyTypeChecker
    return entity_cls


def _add_relation_to_entity_cls_conf(
        conf: EntityClsConf,
        relation,
        relation_name: str,
        relation_field: Relation,
        relation_type: str,
):
    """Adds a relation to an entity class configuration.
    Procedural function that mutates a configuration.

    :param conf: The configuration to add the relation to.
    :param relation: The class resolver of the relation to add.
    :param relation_name: The relation name, meaning the property declared in the relation owner.
    :param relation_field: The relation field to add to the configuration.
    :param relation_type:
    :return: None
    """
    relation_field.entity_cls_resolver = relation
    relation_field.key = _build_foreign_key_name(relation_name=relation_name, relation_type=relation_type)
    conf.relations[relation_name] = relation_field


def _extract_entity_cls_fields(entity_cls) -> List[Tuple]:
    """Extracts the dataclass fields of an entity class.

    :param entity_cls: The entity class to extract the fields from.
    :return: A list of extracted dataclass fields.
    """
    fields = []
    for dataclass_field in dataclass_fields(entity_cls):
        fields.append((
            dataclass_field.name,
            dataclass_field.type,
            field(**_extract_entity_field_attrs(dataclass_field, MISSING))  # pylint: disable=invalid-field-call
        ))

    return fields


def _extract_entity_field_attrs(
        entity_field: Union[DataclassField, Field],
        default_check
) -> Dict:
    """Extracts field attributes from either a dataclass field or a pymnesia field.

    :param entity_field: The object from which to extract the attributes.
    :param default_check: The check value used for checking if a default value or default_factory value has been set.
    :return: A dictionary containing the extracted attributes.
    """
    field_attrs = {}

    if entity_field.default is not default_check:
        field_attrs["default"] = entity_field.default
    if entity_field.default_factory is not default_check:
        field_attrs["default_factory"] = entity_field.default_factory

    return field_attrs


def _add_foreign_key_to_fields(
        relation_name: str,
        fields: List,
        is_nullable: bool,
        relation_type: str,
):
    """Adds a foreign key to an entity class.
    Procedural function that mutates fields.

    :param relation_name: The name of the relation from which the foreign key should be built and added.
    :param fields: The list of fields instance to which the foreign key should be added.
    :param is_nullable: Wether the relation is nullable or not.
    :param relation_type: Wether the relation is nullable or not.
    """
    foreign_key_type = UUID
    if relation_type == "one_to_many":
        foreign_key_type = List[foreign_key_type]
    fields.append((
        _build_foreign_key_name(
            relation_name=relation_name,
            relation_type=relation_type
        ),
        foreign_key_type,
        field(**_build_relation_dataclass_field_args(  # pylint: disable=invalid-field-call
            is_nullable=is_nullable,
            relation_type=relation_type,
        ))
    ))


def _add_relation_to_fields(
        relation_name: str,
        relation: EntityClassResolver,
        fields: List,
        is_nullable: bool,
        relation_type: str,
):
    """Adds a foreign key to an entity class.
    Procedural function that mutates fields.

    :param relation_name: The name of the relation from which the foreign key should be built and added.
    :param relation:
    :param fields: The list of fields instance to which the foreign key should be added.
    :param is_nullable:
    :param relation_type:
    """
    if relation_type == "one_to_many":
        relation = List[relation]
    fields.append(
        (
            relation_name,
            relation,
            field(**_build_relation_dataclass_field_args(  # pylint: disable=invalid-field-call
                is_nullable=is_nullable,
                relation_type=relation_type,
            ))
        )
    )


def _build_relation_dataclass_field_args(
        is_nullable: bool,
        relation_type: str,
) -> dict:
    """Builds the arguments for dataclass field of a relation.

    :param is_nullable: Whether the relation is nullable or not.
    :param relation_type: The type of the relation (one_to_one, one_to_many)
    :return: A dictionary containing the arguments to be used.
    """
    field_args = {}

    default = None if is_nullable and relation_type != "one_to_many" else MISSING
    default_factory = MISSING if relation_type != "one_to_many" else empty_list_factory
    if default is not MISSING:
        field_args["default"] = default
    if default_factory is not MISSING:
        field_args["default_factory"] = default_factory

    return field_args


def _which_one_to_many_relation(typed_attr: Any) -> Union[EntityClassResolver, None]:
    """Determines the entity class resolver from a typed 'one to many' relation.

    :param typed_attr: The typed attribute to determine the entity class resolver from.
    :return: A entity class resolver if the provided typed_attr is a list else None
    """
    if get_origin(typed_attr) == list:
        typing_args = get_args(typed_attr)
        return typing_args[0]

    return None


def _build_foreign_key_name(
        relation_name: str,
        relation_type: str,
) -> str:
    """Builds a foreign key name, based on a relation name.

    :param relation_name: The relation name from which to build the foreign key name.
    :return: A built foreign key name.
    """
    if relation_type == "one_to_many":
        return relation_name[0:-1] + "_ids"
    return relation_name + "_id"


def _validate_entity_cls_declaration(attrs: Dict):
    """Validates that an entity class was correctly required.

    :param attrs: The attributes of the declared entity class.
    :return: None
    """
    if "__tablename__" not in attrs.keys():
        raise MissingTablenameException

    if "id" not in attrs["__annotations__"].keys():
        raise MissingPrimaryKeyException



"""Provides with a metaclass for QueryEngine.
"""
from pymnesia.core.query.engine.base import QueryEngine
from pymnesia.core.query.query import Query
from pymnesia.core.entities.registry.interface import PymnesiaRegistryInterface

__all__ = [
    "QueryEngineType",
    "query_engine_metaclass",
]


class QueryEngineType(type):
    """
    Metaclass for QueryEngine.
    """

    _registry: PymnesiaRegistryInterface = None

    def __new__(mcs, name, bases, attrs):  # pylint: disable=unused-argument
        def table_maker(entity_class_):
            def table_query_maker(self):
                return Query(entity_class=entity_class_, unit_of_work=self.unit_of_work)

            return table_query_maker

        for entity_cls_resolver in mcs._registry.all_configs():
            table = table_maker(entity_cls_resolver)
            table.__name__ = entity_cls_resolver.__tablename__
            attrs[table.__name__] = table

        return super().__new__(mcs, name, (QueryEngine, ), attrs)


def query_engine_metaclass(registry_: PymnesiaRegistryInterface):
    return type(
        QueryEngineType.__name__,
        (QueryEngineType, ),
        {"_registry": registry_}
    )



FILTER_FUNCTIONS_REGISTRY = {}
REL_FILTER_FUNCTIONS_REGISTRY = {}


def register_filter_func(*, filter_name: str, relational: bool = False):
    """Allows to register a filter function

    :param filter_name: The filter name.
    :param relational: Whether the function is used for relational filter or not.
    :return: A decorator that will register the filter function.
    """

    def decorator(func):
        if not relational:
            FILTER_FUNCTIONS_REGISTRY[filter_name] = func
        else:
            REL_FILTER_FUNCTIONS_REGISTRY[filter_name] = func

        return func

    return decorator


def find_filter_function(filter_name: str, relational: bool = False):
    """Retrieves a filter function based on its filter name.

    :param filter_name: The filter name for which to retrieve a filter function.
    :param relational: Whether the function is used for relational filter or not.
    :return: A callable that can be used for filtering.
    """
    if not relational:
        return FILTER_FUNCTIONS_REGISTRY[filter_name]

    return REL_FILTER_FUNCTIONS_REGISTRY[filter_name]
